use std::{collections::HashMap, rc::Rc};

use crate::{
    parse::{Constant, Ident},
    types::TaggedType,
};

use super::DeBrujinAst;

#[derive(Clone)]
pub enum DeBrujinBuiltInAst {
    Abstraction(Ident, Option<TaggedType>, Box<DeBrujinBuiltInAst>), // \:1.2
    Application(Box<DeBrujinBuiltInAst>, Box<DeBrujinBuiltInAst>),   // 0 1
    FreeVariable(String),                                            // x
    BoundVariable(usize),                                            // 1
    Constant(Constant),                                              // true | false | n
    Builtin(Rc<dyn Builtin>),
}

impl DeBrujinAst {
    pub fn resolve_builtins(
        self,
        builtins: &HashMap<String, Rc<dyn Builtin>>,
    ) -> DeBrujinBuiltInAst {
        match self {
            DeBrujinAst::Abstraction(i, t, ast) => {
                DeBrujinBuiltInAst::Abstraction(i, t, Box::new(ast.resolve_builtins(builtins)))
            }
            DeBrujinAst::Application(lhs, rhs) => DeBrujinBuiltInAst::Application(
                Box::new(lhs.resolve_builtins(builtins)),
                Box::new(rhs.resolve_builtins(builtins)),
            ),
            DeBrujinAst::FreeVariable(x) => {
                if let Some(b) = builtins.get(&x) {
                    DeBrujinBuiltInAst::Builtin(b.clone())
                } else {
                    DeBrujinBuiltInAst::FreeVariable(x)
                }
            }
            DeBrujinAst::BoundVariable(b) => DeBrujinBuiltInAst::BoundVariable(b),
            DeBrujinAst::Constant(c) => DeBrujinBuiltInAst::Constant(c),
        }
    }
}

pub trait Builtin {
    fn name(&self) -> String;

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::FreeVariable(self.name())
    }

    fn r#type(&self) -> TaggedType;

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst>;
}

impl DeBrujinBuiltInAst {
    pub(super) fn reduce(self) -> DeBrujinBuiltInAst {
        match self {
            DeBrujinBuiltInAst::Application(lhs, rhs) => {
                let lhs = if lhs.is_value() { *lhs } else { lhs.reduce() };
                let rhs = if rhs.is_value() { *rhs } else { rhs.reduce() };
                assert!(lhs.is_value() && rhs.is_value());
                match lhs {
                    DeBrujinBuiltInAst::Abstraction(_, _, ast) => {
                        let ast = ast.subst_bound(1, rhs);
                        if ast.is_value() { ast } else { ast.reduce() }
                    }
                    DeBrujinBuiltInAst::Builtin(builtin) => match builtin.apply(rhs.clone()) {
                        Some(a) => {
                            if a.is_value() {
                                a
                            } else {
                                a.reduce()
                            }
                        }
                        None => DeBrujinBuiltInAst::Application(
                            Box::new(DeBrujinBuiltInAst::Builtin(builtin)),
                            Box::new(rhs),
                        ),
                    },
                    a => unreachable!(
                        "We cannot get a variable or constant because of type checking and we cannot get a application because that is not a value, {:?}",
                        <DeBrujinBuiltInAst as Into<DeBrujinAst>>::into(a)
                    ),
                }
            }
            a => a,
        }
    }

    fn subst_bound(self, depth: usize, subst: DeBrujinBuiltInAst) -> DeBrujinBuiltInAst {
        match self {
            DeBrujinBuiltInAst::Abstraction(i, t, ast) => {
                DeBrujinBuiltInAst::Abstraction(i, t, Box::new(ast.subst_bound(depth + 1, subst)))
            }
            DeBrujinBuiltInAst::Application(lhs, rhs) => DeBrujinBuiltInAst::Application(
                Box::new(lhs.subst_bound(depth, subst.clone())),
                Box::new(rhs.subst_bound(depth, subst)),
            ),
            DeBrujinBuiltInAst::BoundVariable(n) if n == depth => subst,
            a => a,
        }
    }

    fn is_value(&self) -> bool {
        !matches!(self, DeBrujinBuiltInAst::Application(_, _))
    }
}

impl From<DeBrujinBuiltInAst> for DeBrujinAst {
    fn from(val: DeBrujinBuiltInAst) -> Self {
        match val {
            DeBrujinBuiltInAst::Abstraction(i, t, ast) => {
                DeBrujinAst::Abstraction(i, t, Box::new((*ast).into()))
            }
            DeBrujinBuiltInAst::Application(lhs, rhs) => {
                DeBrujinAst::Application(Box::new((*lhs).into()), Box::new((*rhs).into()))
            }
            DeBrujinBuiltInAst::FreeVariable(x) => DeBrujinAst::FreeVariable(x),
            DeBrujinBuiltInAst::BoundVariable(i) => DeBrujinAst::BoundVariable(i),
            DeBrujinBuiltInAst::Constant(constant) => DeBrujinAst::Constant(constant),
            DeBrujinBuiltInAst::Builtin(builtin) => builtin.to_ast(),
        }
    }
}
