use std::{
    cell::RefCell,
    collections::HashMap,
    fmt::{Debug, DebugMap},
    rc::Rc,
};

use crate::{
    DeBrujinAst,
    multi_map::MultiMap,
    parse::{Constant, Ident},
    types::{PrimitiveType, TaggedType, Type, TypeTag},
    vec_map::VecMap,
};

use super::InferError;

type TypeVar = Type;

pub struct TypeVarCtx {
    counter: RefCell<usize>,
}

impl Default for TypeVarCtx {
    fn default() -> Self {
        Self::new()
    }
}

impl TypeVarCtx {
    pub fn new() -> Self {
        Self {
            counter: RefCell::new(0),
        }
    }

    pub fn get_var(&self) -> Type {
        let mut num = self.counter.borrow_mut();
        let res = format!("?typ_{num}");
        *num += 1;
        Type::Variable(res)
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum TypeVarAst {
    Abstraction(TypeVar, Ident, Type, Box<TypeVarAst>), // \0:1.2
    Application(TypeVar, Box<TypeVarAst>, Box<TypeVarAst>), // 0 1
    FreeVariable(TypeVar, Ident),                       // x
    BoundVariable(TypeVar, usize),                      // 1
    Constant(TypeVar, Constant),                        // true | false | n
}

pub type Constraints<T> = MultiMap<TypeVar, T>;

pub(super) fn step_1(
    ast: DeBrujinAst,
    gamma_free: &HashMap<Ident, TaggedType>,
    mut gamma_bound: Rc<VecMap<usize, Type>>,
    eq_constraints: Rc<RefCell<Constraints<Type>>>,
    st_constraints: Rc<RefCell<Constraints<TypeTag>>>,
    ctx: &TypeVarCtx,
) -> Result<(TypeVarAst, TypeVar), InferError> {
    match ast {
        DeBrujinAst::Abstraction(i, Some(typ), ast) => {
            // let var = ctx.get_var();

            let typ = typ.make_constraints(st_constraints.clone(), ctx);

            let gamma_ref = Rc::make_mut(&mut gamma_bound);
            gamma_ref.map_keys(|i| *i += 1);
            gamma_ref.insert(1, typ.clone());

            let (ast, rhs_var) = step_1(
                *ast,
                gamma_free,
                gamma_bound,
                eq_constraints.clone(),
                st_constraints.clone(),
                ctx,
            )?;
            Ok((
                TypeVarAst::Abstraction(
                    Type::arrow(typ.clone(), rhs_var.clone()),
                    i,
                    typ.clone(),
                    Box::new(ast),
                ),
                Type::arrow(typ, rhs_var.clone()),
            ))
        }
        DeBrujinAst::Abstraction(i, None, ast) => {
            let typ = ctx.get_var();
            let gamma_ref = Rc::make_mut(&mut gamma_bound);
            gamma_ref.map_keys(|i| *i += 1);
            gamma_ref.insert(1, typ.clone());

            let (ast, rhs_var) = step_1(
                *ast,
                gamma_free,
                gamma_bound,
                eq_constraints.clone(),
                st_constraints.clone(),
                ctx,
            )?;
            Ok((
                TypeVarAst::Abstraction(
                    Type::arrow(typ.clone(), rhs_var.clone()),
                    i,
                    typ.clone(),
                    Box::new(ast),
                ),
                Type::arrow(typ, rhs_var.clone()),
            ))
        }
        DeBrujinAst::Application(lhs, rhs) => {
            let var = ctx.get_var();
            let (lhs, lhs_var) = step_1(
                *lhs,
                gamma_free,
                gamma_bound.clone(),
                eq_constraints.clone(),
                st_constraints.clone(),
                ctx,
            )?;
            let (rhs, rhs_var) = step_1(
                *rhs,
                gamma_free,
                gamma_bound.clone(),
                eq_constraints.clone(),
                st_constraints.clone(),
                ctx,
            )?;
            RefCell::borrow_mut(&eq_constraints).insert(lhs_var, Type::arrow(rhs_var, var.clone()));
            Ok((
                TypeVarAst::Application(var.clone(), Box::new(lhs), Box::new(rhs)),
                var,
            ))
        }
        DeBrujinAst::FreeVariable(v) => {
            let var = ctx.get_var();
            let typ = gamma_free
                .get(&v)
                .cloned()
                .ok_or(InferError::NotInContext)?;

            let typ = typ.make_constraints(st_constraints.clone(), ctx);

            RefCell::borrow_mut(&eq_constraints).insert(var.clone(), typ);

            Ok((TypeVarAst::FreeVariable(var.clone(), v), var))
        }
        DeBrujinAst::BoundVariable(i) => {
            let var = ctx.get_var();
            let typ = gamma_bound
                .get(&i)
                .cloned()
                .ok_or(InferError::NotInContext)?;

            RefCell::borrow_mut(&eq_constraints).insert(var.clone(), typ);

            Ok((TypeVarAst::BoundVariable(var.clone(), i), var))
        }
        DeBrujinAst::Constant(constant) => {
            let var = ctx.get_var();
            let typ = match constant {
                Constant::Nat(_) => Type::Primitive(PrimitiveType::Nat),
                Constant::Float(_) => Type::Primitive(PrimitiveType::Float),
                Constant::Bool(_) => Type::Primitive(PrimitiveType::Bool),
            };

            RefCell::borrow_mut(&eq_constraints).insert(var.clone(), typ);

            Ok((TypeVarAst::Constant(var.clone(), constant), var))
        }
    }
}

pub(super) fn step_2(mut eq_constraints: Constraints<Type>) -> Result<Option<SubstFn>, InferError> {
    if let Some((s, t)) = eq_constraints.pop() {
        if s == t {
            step_2(eq_constraints)
        } else if s.type_var().is_some_and(|x| !t.name_used(&x)) {
            let Some(x) = s.type_var() else {
                unreachable!()
            };
            eq_constraints.map(|(k, v)| (k.subst_typevar(&x, &t), v.subst_typevar(&x, &t)));
            let subst = step_2(eq_constraints)?;
            Ok(Some(SubstFn::new(x, t, subst)))
        } else if t.type_var().is_some_and(|x| !s.name_used(&x)) {
            let Some(x) = t.type_var() else {
                unreachable!()
            };
            eq_constraints.map(|(k, v)| (k.subst_typevar(&x, &s), v.subst_typevar(&x, &s)));

            let subst = step_2(eq_constraints)?;
            Ok(Some(SubstFn::new(x, s, subst)))
        } else if let (Some((s_lhs, s_rhs)), Some((t_lhs, t_rhs))) =
            (s.split_arrow(), t.split_arrow())
        {
            eq_constraints.insert(s_lhs, t_lhs);
            eq_constraints.insert(s_rhs, t_rhs);
            step_2(eq_constraints)
        } else {
            Err(InferError::UnusableConstraint)
        }
    } else {
        Ok(None)
    }
}

pub trait Subst {
    fn subst(self, var: &str, subst: &Type) -> Self
    where
        Self: Sized;
}

impl Subst for TypeVarAst {
    fn subst(self, var: &str, subst: &Type) -> TypeVarAst {
        match self {
            TypeVarAst::Abstraction(type1, ident, type2, ast) => TypeVarAst::Abstraction(
                type1.subst_typevar(var, subst),
                ident,
                type2.subst_typevar(var, subst),
                Box::new(ast.subst(var, subst)),
            ),
            TypeVarAst::Application(typ, lhs, rhs) => TypeVarAst::Application(
                typ.subst_typevar(var, subst),
                Box::new(lhs.subst(var, subst)),
                Box::new(rhs.subst(var, subst)),
            ),
            TypeVarAst::FreeVariable(typ, x) => {
                TypeVarAst::FreeVariable(typ.subst_typevar(var, subst), x)
            }
            TypeVarAst::BoundVariable(typ, i) => {
                TypeVarAst::BoundVariable(typ.subst_typevar(var, subst), i)
            }
            TypeVarAst::Constant(typ, constant) => {
                TypeVarAst::Constant(typ.subst_typevar(var, subst), constant)
            }
        }
    }
}

impl Subst for Constraints<TypeTag> {
    fn subst(self, var: &str, subst: &Type) -> Constraints<TypeTag> {
        self.into_iter()
            .map(|(k, v)| (k.subst_typevar(var, subst), v))
            .collect()
    }
}

pub struct SubstFn {
    var: String,
    subst: Type,
    then: Option<Box<Self>>,
}

impl Debug for SubstFn {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut f = f.debug_map();
        self.debug_map(&mut f);
        f.finish()
    }
}

impl SubstFn {
    pub fn new(var: String, subst: Type, then: Option<Self>) -> Self {
        Self {
            var,
            subst,
            then: then.map(Box::new),
        }
    }

    pub fn apply<T: Subst>(&self, target: T) -> T {
        let target = target.subst(&self.var, &self.subst);
        if let Some(s) = &self.then {
            s.apply(target)
        } else {
            target
        }
    }

    fn debug_map(&self, f: &mut DebugMap) {
        f.entry(&self.var, &self.subst);
        if let Some(s) = &self.then {
            s.debug_map(f);
        }
    }
}

pub fn infer_type(
    gamma: &HashMap<Ident, TaggedType>,
    ast: DeBrujinAst,
) -> Result<TaggedType, InferError> {
    let gamma_bound = Rc::new(VecMap::new());
    let eq_constraints = Rc::new(RefCell::new(MultiMap::new()));
    let st_constraints = Rc::new(RefCell::new(MultiMap::new()));
    let ctx = TypeVarCtx::new();

    let (ast, _) = step_1(
        ast,
        gamma,
        gamma_bound,
        eq_constraints.clone(),
        st_constraints.clone(),
        &ctx,
    )?;
    let res = step_2(eq_constraints.take())?.unwrap();

    let ast = res.apply(ast);

    fn get_type(ast: TypeVarAst) -> Type {
        match ast {
            TypeVarAst::Abstraction(_, _, typ, ast) => Type::arrow(typ, get_type(*ast)),
            TypeVarAst::Application(typ, _, _)
            | TypeVarAst::FreeVariable(typ, _)
            | TypeVarAst::BoundVariable(typ, _)
            | TypeVarAst::Constant(typ, _) => typ,
        }
    }

    let typ = get_type(ast);

    let st_constraints = res.apply(st_constraints.take());
    let st_constraints = st_constraints
        .into_iter()
        .map(|(typ, tag)| {
            if typ.has_tag(&tag) {
                Ok((typ, tag))
            } else {
                Err(InferError::DoesNotFitTag(tag, typ.into()))
            }
        })
        .collect::<Result<MultiMap<_, _>, _>>()?;

    let mut typ: TaggedType = typ.into();

    for free_var in typ.free_vars() {
        let tags = st_constraints.get(&Type::Variable(free_var.clone()));
        if tags.is_empty() {
            typ = TaggedType::Tagged(TypeTag::Any, free_var, Box::new(typ));
        } else {
            let tag = tags
                .into_iter()
                .try_fold(TypeTag::Any, |acc, t| acc.tightens(t.clone()))
                .ok_or(InferError::ConfilictingTags)?;
            typ = TaggedType::Tagged(tag, free_var, Box::new(typ));
        }
    }

    Ok(typ)
}

impl TaggedType {
    pub fn make_constraints(
        self,
        st_constraints: Rc<RefCell<Constraints<TypeTag>>>,
        ctx: &TypeVarCtx,
    ) -> Type {
        match self {
            TaggedType::Tagged(type_tag, ident, mut tagged_type) => {
                let var = ctx.get_var();
                let name = var.type_var().unwrap();

                tagged_type.map_name(|n| {
                    if *n == ident {
                        n.clone_from(&name);
                    }
                });

                st_constraints.borrow_mut().insert(var, type_tag);

                tagged_type.make_constraints(st_constraints, ctx)
            }
            TaggedType::Concrete(c) => c,
        }
    }
}

impl Type {
    pub fn subst_typevar(self, var: &str, subst: &Type) -> Type {
        match self {
            Type::Variable(v) if v == var => subst.clone(),
            Type::Arrow(lhs, rhs) => Type::Arrow(
                Box::new(lhs.subst_typevar(var, subst)),
                Box::new(rhs.subst_typevar(var, subst)),
            ),
            t => t,
        }
    }
}
