use std::{collections::HashSet, fmt::Display};

use crate::parse::Ident;

use super::concrete::Type;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[non_exhaustive]
pub enum TypeTag {
    Num,
    Any,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum TaggedType {
    Tagged(TypeTag, Ident, Box<TaggedType>),
    Concrete(Type),
}

impl From<Type> for TaggedType {
    fn from(value: Type) -> Self {
        Self::Concrete(value)
    }
}

impl TypeTag {
    // If one tag tightens the other return that tag, otherwise none,
    pub fn tightens(self, other: Self) -> Option<Self> {
        #[allow(unreachable_patterns)] // #[non_exhaustive] doesn't apply with tuples
        match (self, other) {
            (TypeTag::Any, t) | (t, TypeTag::Any) => Some(t),
            (TypeTag::Num, TypeTag::Num) => Some(Self::Num),
            _ => None,
        }
    }
}

impl TaggedType {
    pub fn into_concrete(self) -> Option<Type> {
        match self {
            TaggedType::Tagged(_, _, _) => None,
            TaggedType::Concrete(t) => Some(t),
        }
    }

    pub fn matches(&self, typ: &Type) -> (bool, Option<Ident>) {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                if typ.has_tag(type_tag) {
                    (true, Some(ident.clone()))
                } else {
                    tagged_type.matches(typ)
                }
            }
            TaggedType::Concrete(c) => (c == typ, None),
        }
    }

    pub fn split_arrow(self) -> Option<(TaggedType, TaggedType)> {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                let (lhs, rhs) = tagged_type.split_arrow()?;
                Some((
                    TaggedType::Tagged(type_tag.clone(), ident.clone(), Box::new(lhs))
                        .clear_unused_names(),
                    TaggedType::Tagged(type_tag, ident, Box::new(rhs)).clear_unused_names(),
                ))
            }
            TaggedType::Concrete(Type::Arrow(lhs, rhs)) => {
                Some((TaggedType::Concrete(*lhs), TaggedType::Concrete(*rhs)))
            }
            TaggedType::Concrete(_) => None,
        }
    }

    pub fn make_arrow(self, rhs: TaggedType) -> Self {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                TaggedType::Tagged(type_tag, ident, Box::new(tagged_type.make_arrow(rhs)))
            }
            TaggedType::Concrete(t) => rhs.make_arrow_lhs(t),
        }
    }

    fn make_arrow_lhs(self, lhs: Type) -> Self {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                TaggedType::Tagged(type_tag, ident, Box::new(tagged_type.make_arrow_lhs(lhs)))
            }
            TaggedType::Concrete(t) => Type::arrow(lhs, t).into(),
        }
    }

    pub fn specialize(self, ident: &str, typ: &Type) -> TaggedType {
        match self {
            TaggedType::Tagged(_, i, tagged_type) if i == ident => {
                tagged_type.specialize(ident, typ)
            }
            TaggedType::Tagged(t, i, tagged_type) => {
                TaggedType::Tagged(t, i, Box::new(tagged_type.specialize(ident, typ)))
            }
            TaggedType::Concrete(c) => TaggedType::Concrete(c.specialize(ident, typ)),
        }
    }

    pub fn map_name<F: Fn(&mut String)>(&mut self, f: F) {
        match self {
            TaggedType::Tagged(_, ident, tagged_type) => {
                f(ident);
                tagged_type.map_name(f);
            }
            TaggedType::Concrete(t) => t.map_name(&f),
        }
    }

    pub fn type_var(&self) -> Option<String> {
        match self {
            TaggedType::Tagged(_, _, tagged_type) => tagged_type.type_var(),
            TaggedType::Concrete(c) => c.type_var(),
        }
    }

    pub fn free_vars(&self) -> HashSet<String> {
        match self {
            TaggedType::Tagged(_, ident, tagged_type) => {
                let mut vars = tagged_type.free_vars();
                vars.retain(|v| v != ident);
                vars
            }
            TaggedType::Concrete(c) => c.type_vars(),
        }
    }

    pub fn name_used(&self, ident: &str) -> bool {
        match self {
            TaggedType::Tagged(_, _, tagged_type) => tagged_type.name_used(ident),
            TaggedType::Concrete(c) => c.name_used(ident),
        }
    }

    pub fn is_concrete(&self) -> bool {
        match self {
            TaggedType::Tagged(_, _, tagged_type) => tagged_type.is_concrete(),
            TaggedType::Concrete(c) => c.is_concrete(),
        }
    }

    fn clear_unused_names(self) -> TaggedType {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                if tagged_type.name_used(&ident) {
                    TaggedType::Tagged(type_tag, ident, Box::new(tagged_type.clear_unused_names()))
                } else {
                    tagged_type.clear_unused_names()
                }
            }
            t @ TaggedType::Concrete(_) => t,
        }
    }
}

impl Display for TypeTag {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            TypeTag::Num => write!(f, "Num"),
            TypeTag::Any => write!(f, "Any"),
        }
    }
}

impl Display for TaggedType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            TaggedType::Tagged(type_tag, ident, tagged_type) => {
                write!(f, "{type_tag} {ident} => {tagged_type}")
            }
            TaggedType::Concrete(t) => write!(f, "{t}"),
        }
    }
}
